/*
 * CG Madness - a Marble Madness clone
 * Copyright (C) 2007  Sven Reinck <sreinck@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "game.hpp"

#include "common.hpp"
#include "level.hpp"
#include "highscore.hpp"

#include "objects.hpp"
#include "ball.hpp"
#include "field.hpp"
#include "menumanager.hpp"
#include "gamemenu.hpp"
#include "features.hpp"
#include "keyboard.hpp"
#include "camera.hpp"
#include "callback.hpp"
#include "text.hpp"
#include "texture.hpp"
#include "environment.hpp"
#include "noise.hpp"
#include "main.hpp"

#include "stringlist.hpp"

#include "color.hpp"

#include "functions.hpp"

#include "types.hpp"

#include <GL/glew.h>
#include <GL/glut.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define FOG_DENSITY 0.003f

static bool gIsGameRunning;

Portal sgPortal1;
Portal sgPortal2;

static float gGameTime;

static const char *gHotSeatLevel = NULL;

void pauseGame(void)
{
	disableBallCamera();
	gIsGameRunning = false;
}

void resumeGame(void)
{
	enableBallCamera();
	gIsGameRunning = true;
}

void setHotSeatLevel(const char *filename)
{
	gHotSeatLevel = filename;
}

void stopWatch(void)
{
	int i;

	int tenthSecond = (int) (gGameTime * 10.0f);
	int newIndex = sgLevel.cntScoreCols;

	while (newIndex > 0 && tenthSecond < sgLevel.scores[newIndex - 1].tenthSecond)
	{
		newIndex--;
	}

	sgLevel.cntScoreCols = min(sgLevel.cntScoreCols + 1, MAX_SCORE_COLS);

	for (i = sgLevel.cntScoreCols - 1; i > newIndex; i--)
	{
		sgLevel.scores[i] = sgLevel.scores[i - 1];
	}

	if (newIndex < MAX_SCORE_COLS)
	{
		sgLevel.scores[newIndex].name[0] = '\0';
		sgLevel.scores[newIndex].tenthSecond = tenthSecond;
	}

	sgLastPlayerIndex = newIndex;
}

void finishedGame()
{
	stopWatch();
	pauseGame();
	showGameMenu(2);
}

void updateGame(float interval)
{
	updateEnvironment(interval);
	if (gIsGameRunning)
	{
		if (wasKeyPressed(KEY_ESC))
		{
			pauseGame();
			showGameMenu(1);
		}

		/* manually switch features */
		if (wasFunctionPressed(1))
		{
			setBallShadow(!useBallShadow());
		}

		if (wasFunctionPressed(2))
		{
			setReflection(!useReflection());
		}

		if (wasFunctionPressed(5))
		{
			pauseGame();
			sgIsMouseControl = !sgIsMouseControl;
			resumeGame();
		}

		if (!sgIsBallInPieces)
		{
			gGameTime += interval;
		}

		updateBall(interval);

		if (sgHasBallHitGoal)
		{
			finishedGame();
		}
	}
	else
	{
#if NOISE_TEXTURE
		if (sgLevel.colorMap == 0 && !sgLevel.waiting)
		{
			sgLevel.colorMap = genTexture();
			colorMapToTexture(sgLevel.colorMap);
			resetBall();
		}
#endif

		updateMenuManager(interval);
	}

	updateGameField();
}

void drawGameHUD(float widthWindow, float heightWindow)
{
	int tenthSecond = (int) (gGameTime * 10.0f);
	float scale = 0.06f;
	float widthDefault = widthStrokeText("x:xx.x") * scale;

	char strTime[10];
	float width;
	float height;

	sprintf(strTime, "%d:%02d.%01d", tenthSecond / 600, tenthSecond / 10 % 60,
					tenthSecond % 10);

	width = widthStrokeText(strTime) * scale;
	height = scale;

	glColor3f(1.0f, 1.0f, 0.0f);

	glPushMatrix();

	glTranslatef((widthWindow - widthDefault) / 2.0f, (heightWindow - height),
							 0.0f);
	glScalef(scale, scale, scale);

	drawStrokeThickText(strTime);

	glPopMatrix();
}

void drawGameWaterReflection(void)
{
	drawGameField(FS_CAM_VIEW);
	drawGameBall();
}

void drawGameBallReflection(void)
{
	drawEnvironment(drawGameWaterReflection);
	drawGameField(FS_BALL_VIEW);
/*
	renderPortals(0, drawGamePortals);
*/
}

void drawGamePortals(void)
{
	drawEnvironment(NULL);
	
	drawGameField(FS_ALL);
	drawGameBall();
}

void drawGame(void)
{
	drawEnvironment(drawGameWaterReflection);

	drawGameField(FS_CAM_VIEW);
	drawGameBall();

	renderPortals(drawGamePortals);
	
	if (!gIsGameRunning)
	{
		drawMenuManager();
	}
}

void resetGameTime(void)
{
	gGameTime = 0.0f;
}

void startGame(void)
{
	sgLevel.lightMap = genTexture();
	lightMapToTexture(sgLevel.lightMap);

	updateTexCoords();

#if (NOISE_TEXTURE)
	updateColorMap();

	pushWaitScreen();

	sgLevel.colorMap = 0;
#endif

	initGameField();

	resetGame();
	
	// init portals

	initPortal(&sgPortal1);
	initPortal(&sgPortal2);

//	sgPortal1.pos = vector3(6.5f, 0.5f, 4.5f);
//	sgPortal1.pos = vector3(6.5f, 4.5f, 0.5f);
	sgPortal1.pos = vector3(5.5f, 3.5f, 0.5f);
//	sgPortal1.orientation = mkQuaternion(-90.0, vector3(0.0f, 0.0f, 1.0f));
	sgPortal1.orientation = mkQuaternion(180.0, vector3(0.0f, 0.0f, 1.0f));
	
//	sgPortal2.pos = vector3(9.0f, 8.5f, 5.0f);
	sgPortal2.pos = vector3(2.0f, 3.5f, 5.0f);
	sgPortal2.orientation = mkQuaternion(90.0, vector3(0.0f, 1.0f, 0.0f));
//	sgPortal2.orientation *= mkQuaternion(180.0, vector3(0.0f, 0.0f, 1.0f));

	sgPortal1.pos += sgLevel.origin;
	sgPortal2.pos += sgLevel.origin;
		
	sgPortal1.other = &sgPortal2;
	sgPortal2.other = &sgPortal1;

	addPortal(&sgPortal1);
	addPortal(&sgPortal2);
}

#if 0
static const char *getNextLevelName(void)
{
	if (gHotSeatLevel)
	{
		if (gInGame)
		{
			return NULL;
		}
		else
		{
			return gHotSeatLevel;
		}
	}
	else
	{
		if (gNextLevelIndex < gLevelNames.count)
		{
			char *name = gLevelNames.strings[gNextLevelIndex];
			gNextLevelIndex++;
			return name;
		}
		else
		{
			gNextLevelIndex = 0;
			return NULL;
		}
	}
}
#endif

void stopGame(void)
{
	glDeleteTextures(1, &sgLevel.lightMap);
#if (NOISE_TEXTURE)
	glDeleteTextures(1, &sgLevel.colorMap);
#endif

	destroyGameField();
}

void resetGame(void)
{
	resetBall();
	resetBallCamera();

	resetGameTime();

	updateGameField();

	pauseGame();
	showGameMenu(0);
}

void initFog(void)
{
	int mode = GL_EXP;
	float density = FOG_DENSITY;
	float color[] = { 1.0f, 1.0f, 1.0f, 1.0f };

	glEnable(GL_FOG);
	glFogiv(GL_FOG_MODE, &mode);
	glFogfv(GL_FOG_DENSITY, &density);
	glFogfv(GL_FOG_COLOR, color);
}

void initGame(void)
{
	resetCamera();

	initObjects();

	initFog();

	/* ball */
	initBall();

	/* menu (must be after ball) */
	initGameMenu();
}
